
 

open Tokens

type pos = int

(* Required by ML-Yacc *)
type svalue = Tokens.svalue
type ('a,'b) token = ('a,'b) Tokens.token
type lexresult  = (svalue,pos) token

val pos = ref 1

fun incline () = pos := (!pos + 1)
fun resetline () = pos := 1

fun eof _ = (let val n = !pos in resetline (); EOF(n,n) end)

val error = fn (e,l : int,_) => print ("line " ^ (Int.toString l) ^
	 ": " ^ e ^ "\n")


%%

%header (functor AntAsmLexFun(structure Tokens : AntAsm_TOKENS));

whitechar       = [\ \t] ;
whitespace      = {whitechar}+ ;

newline         = \n\013 | \n ;

nonnl           = [^\013\n] ;
digit           = [0-9] ;
tokenchar       = [^.,\ \t\n\013:;()\"] ;

%%

\n          => ( incline (); continue() );
";"{nonnl}*\n => ( incline (); continue() );
{digit}+    => ( NUMBER(valOf (Int.fromString yytext),!pos,!pos) );

"("              => ( LPAREN(!pos,!pos) );
")"              => ( RPAREN(!pos,!pos) );
","              => ( COMMA(!pos,!pos) );

"Next"           => ( NEXT(!pos,!pos) );
"."              => ( NEXT(!pos,!pos) );

"Here"           => ( HERE(!pos,!pos) );
"Ahead"          => ( AHEAD(!pos,!pos) );
"LeftAhead"      => ( LEFTAHEAD(!pos,!pos) );
"RightAhead"     => ( RIGHTAHEAD(!pos,!pos) );
"AheadLeft"      => ( LEFTAHEAD(!pos,!pos) );
"AheadRight"     => ( RIGHTAHEAD(!pos,!pos) );
"Left"           => ( LEFT(!pos,!pos) );
"Right"          => ( RIGHT(!pos,!pos) );
"Friend"         => ( FRIEND(!pos,!pos) );
"Foe"            => ( FOE(!pos,!pos) );
"FriendWithFood" => ( FRIENDWITHFOOD(!pos,!pos) );
"FoeWithFood"    => ( FOEWITHFOOD(!pos,!pos) );
"Food"           => ( FOOD(!pos,!pos) );
"Rock"           => ( ROCK(!pos,!pos) );
"Marker"         => ( MARKER(!pos,!pos) );
"FoeMarker"      => ( FOEMARKER(!pos,!pos) );
"Home"           => ( HOME(!pos,!pos) );
"FoeHome"        => ( FOEHOME(!pos,!pos) );

"AllOf"          => ( ALLOF(!pos,!pos) );
"AnyOf"         => ( SOMEOF(!pos,!pos) );
"Not"            => ( NOT(!pos,!pos) );
"True"           => ( TRUE(!pos,!pos) );
"False"          => ( FALSE(!pos,!pos) );

"Sensing"        => ( SENSING(!pos,!pos) );

"Sense"          => ( SENSE(!pos,!pos) );
"Mark"           => ( MARK(!pos,!pos) );
"Unmark"         => ( UNMARK(!pos,!pos) );
"PickUp"         => ( PICKUP(!pos,!pos) );
"Drop"           => ( DROP(!pos,!pos) );
"Turn"           => ( TURN(!pos,!pos) );
"Move"           => ( MOVE(!pos,!pos) );
"Flip"           => ( FLIP(!pos,!pos) );
":"              => ( COLON(!pos,!pos) );

"Func"           => ( FUNC(!pos,!pos) );
"Call"           => ( CALL(!pos,!pos) );
"Print"          => ( PRINT(!pos,!pos) );
"Abort"          => ( ABORT(!pos,!pos) );
"SetName"        => ( SETNAME(!pos,!pos) );
"Include"        => ( INCLUDE(!pos,!pos) );
"Define"         => ( DEFINE(!pos,!pos) );

"If"             => ( IF(!pos,!pos) );
"Then"           => ( THEN(!pos,!pos) );
"Else"           => ( ELSE(!pos,!pos) );
"End"            => ( END(!pos,!pos) );
"Goto"           => ( GOTO(!pos,!pos) );
"Repeat"         => ( REPEAT(!pos,!pos) );
"While"          => ( WHILE(!pos,!pos) );

"Variable"       => ( VARIABLE(!pos,!pos) );
"Of"             => ( OF(!pos,!pos) );
"In"             => ( IN(!pos,!pos) );

"Switch"         => ( SWITCH(!pos,!pos) );
"Case"           => ( CASE(!pos,!pos) );

"Assign"         => ( ASSIGN(!pos,!pos) );
"Set"            => ( ASSIGN(!pos,!pos) );

{tokenchar}+     => ( IDENT(yytext,!pos,!pos) );
{whitespace}     => ( continue() );

\"[^\"]*\"  => ( STRING(String.substring(yytext,1,(String.size yytext)-2),!pos,!pos) );

