(* asmel.sml *)
(* Team Tycon Mismatch, ICFP 04 Programming Contest *)


structure AsmEL = 
  struct
    type label = string
    datatype state = Next | Label of label

    datatype sensedir = Here | Ahead | LeftAhead | RightAhead
    type marker = int (* ?? *)
    datatype lor = Left | Right
    datatype sensecond = 
      Friend | Foe | FriendWithFood | FoeWithFood | Food | Rock | Marker of expr | FoeMarker | Home | FoeHome
    (* Cooler kinds of conditions *)
      | sAllOf of sensecond list
      | sSomeOf of sensecond list
      | sNot of sensecond

    and cond =
      Sensing of expr * expr
    | Flipped of expr
    | AllOf of cond list
    | SomeOf of cond list
    | Not of cond
    | True
    | False

    and expr = 
      SenseDir of sensedir
    | SenseCond of sensecond
    | Number of int
    | FlipProp of int
    | LeftOrRight of lor
    | Cond of cond
    | Var of string
    | eNext
    | eAllOf of expr list
    | eSomeOf of expr list
    | eNot of expr

    datatype instruction = 
      Sense of expr * expr * expr * expr
    | Mark of expr * expr
    | Unmark of expr * expr
    | PickUp of expr * expr
    | Drop of expr
    | Turn of expr * expr
    | Move of expr * expr
    | Flip of expr * expr * expr
    | LabelDef of string
    (* Non-standard *)
    | Print of string
    | Abort of string
    | SetName of char * char
    (* Compound *)
    | IfGoto of expr * expr
    | IfThenElse of expr * instruction list * instruction list
    | StateVar of string * string list * instruction list
    | Switch of string * (string * (instruction list)) list
    | StateAssign of int ref * string * string  (* int ref will hold a unique identifier for this assignment statement. *)
    | While of expr * instruction list
    (* Macro *)
    | Func of string * string list * instruction list
    | Call of string * expr list
    | Repeat of expr * instruction list
    | Include of string
    | Define of string * expr

    type antbrain = instruction Vector.vector

(* Not currently used.
    fun outExpr out i =
        (case i of
           SenseDir _ => ()
         | SenseCond _ => ()
         | Number _ => ()
         | FlipProp _ => ()
         | LeftOrRight _ => ()
         | Cond _ => ()
         | Var v => out v
         | eNext => out "Next")

    fun outInstruction out i =
        (case i of
           Sense (sd, st1, st2, cond) => ()
         | Mark (m, st) => ()
         | Unmark (m, st) => ()
         | PickUp (st1, st2) => ()
         | Drop st => ()
         | Turn (lor, st) => ()
         | Move (st1, st2) => ()
         | Flip (p, st1, st2) => ()
         | LabelDef l => out (l ^ ":\n")
         | Print s => ()
         | Abort s => ()
         | SetName (c1, c2) => ()
         | IfGoto (c, st) => ()
         | IfThenElse (c, insts1, insts2) => ()
         | StateVar (name, choices, insts) => ()
         | Switch (var, arms) => ()
         | StateAssign (_, var, newval) => ()
         | While _ => ()
         | Func (f, args, insts) => ()
         | Call (f, args) => 
           (out ("Call " ^ f ^ " (");
            List.app (fn arg => ((outExpr out arg); out ", ")) args;
            out ")\n")
	 | Repeat _ => ()
         | Include _ => ()
         | Define _ => ())
*)

  end

