
structure Optimize :> OPTIMIZE =
struct

  val opt = Params.flag false
                          (SOME ("-O", 
                                 "optimize the instructions before output")) "opt"

  open AsmIL

  exception Optimize of string

  fun insmap f i =
    (case i of
       Sense (sd, a, b, cond) => Sense (sd, f a, f b, cond)
     | Mark (m, a) => Mark(m, f a)
     | Unmark (m, a) => Unmark (m, f a)
     | PickUp (a, b) => PickUp (f a, f b)
     | Drop a => Drop (f a)
     | Turn (lor, a) => Turn (lor, f a)
     | Move (a, b) => Move (f a, f b)
     | Flip (i, a, b) => Flip (i, f a, f b)
     | Abort s => Abort s
     | Print s => Print s
     | SetName x => SetName x)

  (* just path compression for now *)
  fun optimize orig =
    if !opt then
       let
         fun go 0 ooo = ooo
           | go n ooo =
           let
             val news =
               Array.tabulate(Vector.length ooo, fn x => x)
               
             fun set x a = 
               let in
                 (* print ("optimize(" ^ Int.toString n ^ 
                    "): state " ^ Int.toString x ^ " to " ^
                    Int.toString a ^ "\n"); *)
                 Array.update(news, x, a)
               end
             
             fun direct x ins =
               (case ins of
                  Sense (_, a, b, _) => if a = b then set x a else ()
                | Flip (_, a, b) => if a = b then set x a else ()
                | _ => ())
                  
           in
             Vector.appi (fn (idx, ins) => direct idx ins) ooo;
             
             
             (* perform mapping *)
             go (n - 1) (Vector.map (insmap (fn x =>
                                             Array.sub(news, x))) ooo)
           end
       in
         go 50 orig
       end
     else orig
end
