;  Kicker is a variation of Tom's attack ants.  It can either attack the
;  enemy hill or defend your own.  It accomplishes both by
;  running around the outside of the respective hill.

;;; Create an attacker ant.  Given a label that defines what to do if there 
;;; are already enough attackers
;Func KICKER_ATTACKER (rtrn)

;;; Create an attacker ant.  Given a label that defines what to do if there 
;;; is food just outside of the anthill and one if there is already enough
;;; defenders.  Note that rtrnfood should NOT drop food for friends (otherwise
;;; other defenders will jump out and pick it up).
;Func KICKER_DEFENDER (rtrnfood, rtrnnofood)

Func _KICKER (rtrnfood, rtrnnofood, whch)
Variable mode Of (defend attack) In
;;    Flip 3 attack forage
;;;  this time we always start with attack
    If whch Then
	Assign mode defend
    Else
	Assign mode attack
    End
	
attack:
    Switch mode Of
    Case defend
	    Sense found-base no-enemy-base Here Home
    Case attack
	    Sense found-base no-enemy-base Here FoeHome
    End	

;;;  try to orient ourselves so that we are on the edge,
;;;  then walk around it until we are adjacent to a friend
found-base:
;;;  might as well pick up their food--it's a free point!
;;;  they probably leave it on the edge, anyway. once we get 
;;;  critical density, we could perhaps start removing it from
;;;  base...
    Switch mode Of
    Case attack
	    PickUp . .
    End

Func TRY_TO_FACE_ME (rtrn)
    If AllOf(Sensing Ahead Home
	     Not (Sensing LeftAhead Home)) Goto rtrn
;;;  not yet, so turn and try again
    Turn Right .
End

try-to-face-cw:
    Switch mode Of
    Case defend
	Call TRY_TO_FACE_ME (at-base-facing-cw)
	Call TRY_TO_FACE_ME (at-base-facing-cw)
	Call TRY_TO_FACE_ME (at-base-facing-cw)
	Call TRY_TO_FACE_ME (at-base-facing-cw)			
;;;  now try to face cw.
	If AllOf(Sensing Ahead Home
	    Not (Sensing LeftAhead Home)) Goto at-base-facing-cw
;;;  not yet, so move
	    Move rtrnnofood blocked-at-base

    Case attack
;;;  now try to face cw.
	If AllOf(Sensing Ahead FoeHome
	     Not (Sensing LeftAhead FoeHome)) Goto at-base-facing-cw
;;;  not yet, so turn and try again
	Turn Right try-to-face-cw

    End
	
;;; walk around the base, to avoid being killed, and to make space
;;; for friends trying to get in
at-base-facing-cw:
    Switch mode Of
    Case defend
	If Sensing LeftAhead Food Goto sneak-away-and-get-food
    End   
    Move found-base at-base-facing-cw

sneak-away-and-get-food:
    Turn Left Next
    Move Next attack
    PickUp . attack
    Turn Left Next
    Turn Left Next
    Turn Left rtrnfood
    ;;  try to find base

    
;;;  keep looking ... 
no-enemy-base:
    Move attack hit-while-attacking

hit-while-attacking:
;;;  see what I hit. if rock, turn:
    If Sensing Ahead Rock Goto turn-away
;;;  if a friend or foe on base,
;;;  just wait for it to move out of the way.
;;;  we have priority because of a tight loop.
;;;  on the other hand, we may want to sense around
;;;  for lots of foes and run if we're about to
;;;  be (accidentally) captured
    Switch mode Of
    Case defend
	Sense blocked-at-base Next Ahead Home
    Case attack
	Sense blocked-at-base Next Ahead FoeHome
    End
;;;  otherwise, just turn away:	
    
turn-away:
    Flip 2 arot-cw arot-ccw
arot-cw:
    Turn Right Next
    Flip 3 arot-cw no-enemy-base
arot-ccw:
    Turn Left Next
    Flip 3 arot-ccw no-enemy-base

;;; I know I'm facing their base, on non-base
blocked-at-base:
;;;  XX randomly decide to go forage. (interval 18 * 30)
    If Flip 540 Goto rtrnnofood
    If AnyOf (Sensing Ahead Foe Sensing Ahead Friend) Goto blocked-at-base
;;;  XX by invt can go directly to found-base
    Move attack blocked-at-base

End
End

Func KICKER_DEFENDER (rtrnfood, rtrnnofood)
    Call _KICKER (rtrnfood, rtrnnofood, True)
End

Func KICKER_ATTACKER (rtrn)
    Call _KICKER (rtrn, rtrn, False)
End
