


structure Combine =
  struct

    open AntGenes
    structure R = Random

    fun fixinst fix inst = 
      (case inst of
	 Sense (d,s,s',c) => Sense (d,fix s,fix s',c)
       | Mark (m,s) => Mark (m,fix s)
       | Unmark (m,s) => Unmark (m,fix s)
       | PickUp (s,s') => PickUp (fix s,fix s')
       | Drop s => Drop (fix s)
       | Turn (lr,s) => Turn (lr,fix s)
       | Move (s,s') => Move (fix s, fix s')
       | Flip (i,s,s') => Flip (i,fix s,fix s'))

    fun shift offset = fixinst (fn srcj => srcj + offset)

    fun pad v len = 
        let 
	  val vlen = Vector.length v
	  fun ithinst i = 
	    let val srci = i mod vlen
	        val fixer = (fn j => (j + (i-srci)) mod len)
	    in fixinst fixer (Vector.sub (v,srci))
	    end
	in Vector.tabulate (len,ithinst)
	end      

    fun splice blocksize (g1,g2) = 
      let 
	val (g1,g2,size) =  (* Make sure they're the same length. *) 
	  let 
	    val n1 = Vector.length g1
	    val n2 = Vector.length g2
	  in
	    if n1=n2 then (g1,g2,n1)
	    else if n1 < n2 then (pad g1 n2,g2,n2)
		 else (g1, pad g2 n1,n1)
	  end
	val dummyinst = Drop 0
	val buf = Array.array (size,dummyinst)
	fun looper (n,which) =
	  if n = size then ()
	  else 
	    let 
	      val which = 
		if n mod blocksize <> 0 then which 
		else (R.randomInt 1 = 0)
	      val copyfrom = if which then g1 else g2
	    in
	      Array.update (buf,n,Vector.sub (copyfrom,n));
	      looper (n+1,which)
	    end
      in looper (0,false);
	Vector.tabulate (size,fn i => Array.sub (buf,i))
      end	


    fun mutateinst (statemutateprobinv,markmutateprobinv,condmutateprobinv,dirmutateprobinv,
		    instmutateprobinv) len inst =
      let 
	fun mutateint (max,probinv) s = if R.randomInt statemutateprobinv = 0 then R.randomInt len else s
	val mutstate = mutateint (len,statemutateprobinv)
	val mutmark  = mutateint (len,markmutateprobinv)

	fun randomDir () = 
	  (case R.randomInt 3 of
	     0 => LeftAhead
	  | 1 => RightAhead
	  | _ => Ahead)

	fun mutsensedir sd = 
	  (if R.randomInt dirmutateprobinv <> 0 then sd
	   else randomDir ())

	fun randomCond () = 
	    (case R.randomInt 10 of
	       0 => Friend
	     | 1 => Foe
	     | 2 => FriendWithFood
	     | 3 => FoeWithFood
	     | 4 => Food
	     | 5 => Rock
	     | 6 => Marker (R.randomInt 6)
	     | 7 => FoeMarker
	     | 8 => Home
	     | _ => FoeHome)

	fun mutcond c =
	  if R.randomInt condmutateprobinv <> 0 then 
	    (case c of
	       Marker m => Marker (mutmark m)
	     | c => c)
	  else randomCond ()

	fun randomState () = R.randomInt len

	fun randomInst () =
	  (case R.randomInt 8 of
	     0 => Sense (randomDir (),R.randomInt len,R.randomInt len,randomCond ())
	   | 1 => Mark (R.randomInt 6,R.randomInt len)
	   | 2 => Unmark (R.randomInt 6,randomState ())
	   | 3 => PickUp (randomState(),randomState ())
	   | 4 => Drop (randomState ())
	   | 5 => Turn (case R.randomInt 2 of 0 => Left | _ => Right,randomState ())
	   | 6 => Move (randomState (),randomState ())
	   | _ => Flip (R.randomInt 1000,randomState (),randomState ()))

      in
	if R.randomInt instmutateprobinv <> 0 then
	  case inst of
	    Sense (d,s,s',c) => Sense (mutsensedir d,mutstate s,mutstate s',mutcond c)
	  | Mark (m,s) => Mark (mutmark m,mutstate s)
	  | Unmark (m,s) => Unmark (mutmark m,mutstate s)
	  | PickUp (s,s') => PickUp (mutstate s,mutstate s')
	  | Drop s => Drop (mutstate s)
	  | Turn (lr,s) => Turn (lr,mutstate s)
	  | Move (s,s') => Move (mutstate s,mutstate s')
	  | Flip (i,s,s') => Flip (i,mutstate s,mutstate s')
	else
	  randomInst ()
      end

    fun mutate params g =
      let val size = Vector.length g
      in Vector.tabulate (size,fn i => mutateinst params size (Vector.sub (g,i)))
      end

  end