

structure GeneMain = 
  struct

    val _ = Random.autoseed ()

    structure C = Combine

    fun splice_files (minsize,blocksize,mutateparams) (f1,f2) outf = 
      let 
	val g1 = C.pad (Parse.parse_file f1) minsize
	val g2 = C.pad (Parse.parse_file f2) minsize
	val outg = C.splice blocksize (g1,g2)
	val outg = C.mutate mutateparams outg
      in Output.output_file outf outg
      end

    val minsize = ref 8000
    val blocksize = ref 100
    val statemutateprob = ref 1000
    val markmutateprob = ref 500
    val condmutateprob = ref 500
    val dirmutateprob  = ref 500
    val instmutateprob = ref 500

    val quicksplice = splice_files (!minsize,
				    !blocksize,
				    (!statemutateprob,
				     !markmutateprob,
				     !condmutateprob,
				     !dirmutateprob,
				     !instmutateprob))

    fun doiter ants childdir listout = 
      let
	val popsize = length ants
	val (mother,father) = 
	  let
	    fun looper () = 
	      let val (m,f) = (Random.randomInt popsize,Random.randomInt popsize)
	      in if m = f then looper ()
		 else (List.nth (ants,m),List.nth (ants,f))
	      end
	  in looper ()
	  end

	val _ = print ("Generating the abominable mutant offspring of " ^ mother ^ " and " ^ father ^ "\n")

	fun freshantname () = (childdir ^ "/" ^ "mutant-" ^ (Int.toString (Random.randomInt (1000000))) ^ ".ant")
	val children = List.tabulate (2,fn i => freshantname ())
	val _ = map (quicksplice (mother,father)) children

	val _ = List.app (fn s => (print "\t";
				   print s;
				   print "\n")) children

	val newpop = children @ ants
	val maxpopsize = 30
	val newpop = if length newpop <= maxpopsize then newpop 
		     else List.take (newpop,maxpopsize)
	
	val outs = TextIO.openOut listout
      in
	List.app (fn s => (TextIO.output (outs,s); TextIO.output (outs,"\n"))) newpop;
	TextIO.closeOut outs
      end

    fun readAntList fname =
      let val ins = TextIO.openIn fname
	  fun loop acc =
	    (case TextIO.inputLine ins of
	       NONE => rev acc
	     | SOME s => loop ( (String.substring (s,0,String.size s - 1)) ::acc))
	  val answer = loop []
	  val _ = TextIO.closeIn ins
      in answer
      end 

  end


val _ = 
  (case Params.docommandline() of
     [ants,childdir,listout] => GeneMain.doiter (GeneMain.readAntList ants) childdir listout
   | _ => print "Usage: genetics <antlist> <directory for children> <output file>\n")