signature ANTS =
sig

    exception Ants of string

    type state = int
    type mark = int
    type dir = int

    datatype color =
        Red 
      | Black

    val othercolor : color -> color

    datatype lr =
        Left
      | Right

    datatype sensedir = 
        Here
      | Ahead
      | LeftAhead
      | RightAhead
    
    datatype cond =
        Friend
      | Foe
      | FriendWithFood
      | FoeWithFood
      | Food
      | Rock
      | Marker of mark
      | FoeMarker
      | Home
      | FoeHome

    datatype inst = 
        Sense of sensedir * state * state * cond
      | Mark of mark * state
      | Unmark of mark * state
      | PickUp of state * state
      | Drop of state
      | Turn of lr * state
      | Move of state * state
      | Flip of int * state * state
      | Print of string
      | Abort of string
      | SetName of char * char

    type brain = inst Vector.vector
    type pos
     
    val mkpos : int * int -> pos
    val unpos : pos -> int * int
      
    type ant = { pos : pos ref,
                 state : int ref,
                 color : color,
                 resting : int ref,
                 dir : dir,
                 names : (char * char) ref,
                 has_food : bool }

    type marks
    val mark_empty : marks
    val mark_set : marks * (color * mark) -> marks
    val mark_clear : marks * (color * mark) -> marks
    val mark_test : marks * (color * mark) -> bool
    val mark_test_any : marks * color -> bool
    val mark_to_char : marks -> char
                                         
    datatype cell =
        Rocky
      | Clear of { ant : int option ref,
                   food : int,
                   hill : color option,
                   marks : marks }

    val turn : lr * dir -> dir
    val next_cell : pos * dir -> pos
    val sensed_cell : pos * dir * sensedir -> pos

    type map

    val initial_map : int * int * cell list list -> map

    (* returns width, height *)
    val get_dimensions : map -> int * int
    val get_cell : map * pos -> cell
    val set_cell : map * pos * cell -> unit

    val set_cell_ant : map -> pos -> int option -> unit
    val get_cell_ant : map -> pos -> int option
    val set_cell_food : map -> pos -> int -> unit
    val get_cell_food : map -> pos -> int
    val set_cell_marks : map -> pos -> marks -> unit
    val get_cell_marks : map -> pos -> marks

end
