structure Ants :> ANTS =
struct

    exception Ants of string

    type state = int
    type mark = int
    type dir = int

    datatype color =
        Red 
      | Black

    fun othercolor Red = Black
      | othercolor Black = Red

    datatype lr =
        Left
      | Right

    datatype sensedir = 
        Here
      | Ahead
      | LeftAhead
      | RightAhead
    
    datatype cond =
        Friend
      | Foe
      | FriendWithFood
      | FoeWithFood
      | Food
      | Rock
      | Marker of mark
      | FoeMarker
      | Home
      | FoeHome

    datatype inst = 
        Sense of sensedir * state * state * cond
      | Mark of mark * state
      | Unmark of mark * state
      | PickUp of state * state
      | Drop of state
      | Turn of lr * state
      | Move of state * state
      | Flip of int * state * state
      | Print of string
      | Abort of string
      | SetName of char * char

    type brain = inst Vector.vector

    (* as y * 1024 + x *)
    type pos = int

    type ant = { pos : pos ref,
                 state : int ref,
                 color : color,
                 resting : int ref,
                 dir : dir,
                 names : (char * char) ref,
                 has_food : bool }

    fun mkpos (x, y) = y * 1024 + x

    fun unpos x = raise Ants "no unpos"

    local
        open Word32
    in
        type marks = word
        fun fromColormark (Red, m) = Word.fromInt m
          | fromColormark (Black, m) = Word.+(Word.fromInt 6, Word.fromInt m)
        fun mark_set (ms, cm) = orb(ms, <<(fromInt 1, fromColormark cm)) 
        fun mark_clear(ms, cm) = andb(ms, notb(<<(fromInt 1, fromColormark cm)))
        fun mark_test(ms, cm) = not(toInt(andb(ms, <<(fromInt 1, fromColormark cm))) = 0)
        fun mark_test_any (ms, Red) = not(toInt(andb(ms, fromInt 63)) = 0)
          | mark_test_any (ms, Black) = not(toInt(andb(ms, fromInt 4032)) = 0)
        val mark_empty = fromInt 0
        fun mark_to_char 0w0 = #" "
          | mark_to_char m = 
          let val m = orb(andb(m, 0w63), andb(>>(m, 0w6), 0w63))
          in
            chr (toInt (andb(m + (fromInt (ord #" ")), 0wxFF)))
          end
    end

    datatype cell =
        Rocky
      | Clear of { ant : int option ref,
                   food : int,
                   hill : color option,
                   marks : marks }

    (* val next_cell : pos * dir -> pos 
     returns the adjacent cell in the given direction *)

    fun even m = 
      Word32.andb(Word32.fromInt m, 0w1) = 0w0

    fun yp n = n div 1024
    fun xp n = n mod 1024

    fun next_cell (n,d) = 
      let
        val x = xp n
        val y = yp n
      in
        case (d, even y) of 
          (0, _) => mkpos (x+1,y)
          | (1, true) => mkpos (x,y+1)
          | (1, false) => mkpos (x+1,y+1)
          | (2, true) => mkpos (x-1,y+1)
          | (2, false) => mkpos (x,y+1)
          | (3, _) => mkpos (x-1,y)
          | (4, true) => mkpos (x-1,y-1)
          | (4, false) => mkpos (x,y-1)
          | (5, true) => mkpos (x,y-1)
          | (5, false) => mkpos (x+1,y-1)
          | _ => raise Ants "bad direction"
      end
    (* val sensed_cell : pos * dir * sensedir -> pos
     returns position of sensed cell *)

    (*  val turn : lr * dir -> dir *)
    fun mod6 d = if d >= 6 then d - 6 else d
(*
    fun turn (Left, d) = (d + 5) mod 6
      | turn (Right, d) = (d + 1) mod 6
*)
    fun turn (Left, d) = mod6 (d + 5)
      | turn (Right, d) = mod6 (d + 1)

    fun sensed_cell (p, d, Here) = p
      | sensed_cell (p, d, Ahead) = next_cell (p, d)
      | sensed_cell (p, d, LeftAhead) = next_cell (p, turn(Left, d))
      | sensed_cell (p, d, RightAhead) = next_cell (p, turn(Right, d))


    (* implement me *)
    type map = { w : int, h : int, data : cell Array.array }

      
    fun get_dimensions { w, h, data } = (w, h)
    fun get_cell ({ w, h, data }, n) =
      Array.sub(data, (yp n) * w + (xp n))
    fun set_cell ({ w, h, data }, n, c) = 
      Array.update(data, (yp n) * w + (xp n), c)


    fun initial_map (w, h, cells) = 
      if w <= 0 orelse h <= 0 
         orelse length cells <> h
         orelse List.exists (fn l => length l <> w) cells
      then raise Ants "bad dimensions to initial_map"
      else 
        let val m = 
          { w = w, h = h,
            data = Array.array (w * h, Rocky) }
        in
          (ListUtil.appi 
           (fn (l : cell list, y) =>
            ListUtil.appi (fn (c, x) =>
                           set_cell (m, mkpos (x, y), c)) l) cells);
          m
        end

    fun set_cell_ant m p i = 
        let
            val c = get_cell (m, p)
        in
            case c of
                Rocky => raise Ants "impossible"
              | Clear { ant, ... } => ant := i
        end

    fun get_cell_ant m p = 
        let
            val c = get_cell (m, p)
        in
            case c of
               Clear { ant = ref ant, food, hill, marks } => ant
             | _ => NONE
        end

    fun set_cell_food m p f = 
        let
            val c = get_cell (m, p)
        in
            case c of
                Rocky => raise Ants "impossible"
              | Clear { ant, food, hill, marks } => 
                    set_cell (m, p, Clear { ant=ant, food=f, hill=hill, marks=marks })
        end

    fun get_cell_food m p = 
        let
            val c = get_cell (m, p)
        in
            case c of
               Clear { food, ... } => food
             | _ => raise Ants "impossible"
        end

    fun set_cell_marks m p mk = 
        let
            val c = get_cell (m, p)
        in
            case c of
                Rocky => raise Ants "impossible"
              | Clear { ant, food, hill, marks } => 
                    set_cell (m, p, Clear { ant=ant, food=food, hill=hill, marks=mk })
        end

    fun get_cell_marks m p = 
        let
            val c = get_cell (m, p)
        in
            case c of
               Clear { marks, ... } => marks
             | _ => raise Ants "impossible"
        end


end