structure Simulator :> SIMULATOR =
struct

    open Ants
    exception Unimp

    (* val mapfold : (pos * cell * 'a -> 'a) -> 'a -> map -> 'a
     calls a function for every cell in the map, in
     left-to-right top-to-bottom order, threading an argument
     through, fold-style. *)

    fun mapfold f a m = 
        let
            val (xs, ys) = get_dimensions m
        in
            Util.ford 0 (ys-1) a 
            (fn (y, a) =>
             Util.ford 0 (xs-1) a 
             (fn (x, a) => f (mkpos (x, y), get_cell (m, mkpos (x, y)), a)))
        end

    (* val populate : map -> antvector *)
    fun populate m = 
        let
            fun make_ant (xy, c as Clear {hill = SOME col, ant, food, marks}, (ants, id)) =
                let 
                    val newant = { pos = ref xy,
                                   state = ref 0,
                                   color = col, 
                                   resting = ref 0, 
                                   dir = 0, 
                                   names = ref (#"X", #"O"),
                                   has_food = false}
                in
                    set_cell (m, xy, Clear { hill = SOME col, 
                                             ant = ref (SOME id), 
                                             food = food, marks=marks});
                    (newant::ants, id+1)
                end
              | make_ant (_, _, info) = info
        in
            Array.fromList (map SOME (rev (#1 (mapfold make_ant ([],0) m))))
        end

    (* val score : map -> int * int *)
    fun score m = 
        let
            fun tally (_, Clear {hill = SOME Red, food = f, ...}, (rs, bs)) = (rs + f, bs)
              | tally (_, Clear {hill = SOME Black, food = f, ...}, (rs, bs)) = (rs, bs + f)
              | tally (_, _, s) = s
        in
            mapfold tally (0,0) m
        end


    fun colortostring Red = "red"
      | colortostring Black = "black"

    fun anttostring (_, NONE) = ""
      | anttostring (ants, SOME i) = 
        (case Array.sub (ants, i) of
             NONE => ""
           | SOME ({ pos = ref pos, state = ref state, color, 
                     resting = ref resting, dir, has_food, ... } : Ants.ant) =>
                 (colortostring color) ^ " ant of id " ^ (Int.toString i) ^ ", dir " ^
                 (Int.toString dir) ^ 
                 ", food " ^ (if has_food then "1" else "0") ^ ", state " ^
                 (Int.toString state) ^ ", resting " ^ (Int.toString resting))

    fun colmarkstostring c m = 
        let
            val s = Util.ford 0 5 "" 
                (fn (i, s) =>
                 s ^ (if mark_test (m, (c, i)) then Int.toString i else ""))
        in
            if s = "" then "" else (colortostring c) ^ " marks: " ^ s ^ "; "
        end

    fun markstostring m = (colmarkstostring Red m) ^ (colmarkstostring Black m)
    fun celltostring (_, Rocky) = "rock"
      | celltostring (ants, Clear {ant = ref ant, food, hill, marks}) = 
             (if food > 0 then (Int.toString food) ^ " food; " else "") ^
             (case hill of NONE => "" | SOME Red => "red hill; " | SOME Black => "black hill; ") ^
             (markstostring marks) ^
             (anttostring (ants, ant))



    fun print_map (n, ants, m) =
        let 
            val (xs, ys) = get_dimensions m
            val _ = print ("\nAfter round " ^ (Int.toString n) ^ "...\n")
            val _ =     Util.for 0 (ys-1) 
                (fn y =>
                 Util.for 0 (xs-1)  
                 (fn x => (print ("cell (" ^ (Int.toString x) ^ ", " ^ (Int.toString y) ^ "): "); 
                           print (celltostring (ants, get_cell (m, mkpos (x,y))));
                           print "\n"
                           )))
        in
            ()
        end
    
    fun run (b1f, b2f, mf, seed, maxround, debug) =
        let
            val _ = Random.seed seed
            val b1 = Parse.parsebrain b1f
            val b2 = Parse.parsebrain b2f
            val m = Parse.parsemap mf
            val ants = populate m 
            val _ = Util.for 0 maxround 
              (fn n => (if debug then print_map (n, ants, m) else ();
                          Step.round (b1, b2, m, ants) n))
        in
            score m
        end

    fun simulate arg =
        let
            val (rs, bs) = run arg
        in
           print ("Red score: " ^ (Int.toString rs) ^  "\n");
           print ("Black score: " ^ (Int.toString bs) ^  "\n")
        end
end
