
structure Step :> STEP =
struct

  open Ants

  exception Step of string

  val vis = Params.flag true
        (SOME ("-vis", 
               "Visualize every step")) "vis"

  val freq = Params.param "1"
        (SOME ("-freq",
               "Only show 1/freq rounds")) "freq"

  val trace = Params.paramacc nil (SOME("-trace", "list of robot ids", #",")) "trace"
  
  fun round (redb, blackb, map, ants) roundnum =
    let

      fun inst Black s = Vector.sub(blackb, s)
        | inst Red s = Vector.sub(redb, s)

      fun cell_matches (pos, cond, color) = 
          let
              val c = get_cell (map, pos)

              fun is_ant (NONE, _) = false
                | is_ant (SOME i, c) = (case Array.sub (ants, i) of
                                          NONE => raise Step "impossible"
                                        | SOME {color, ...} => color = c)

              fun is_ant_wf (NONE, _) = false
                | is_ant_wf (SOME i, c) = (case Array.sub (ants, i) of
                                             NONE => raise Step "impossible2"
                                           | SOME {color, has_food, ...} => 
                                               color = c andalso has_food)
          in
              case c of
                  Rocky => cond = Rock
                | Clear { ant = ref ant, food, hill, marks } => 
                      (case cond of 
                           Friend => is_ant(ant, color)
                         | Foe => is_ant(ant, othercolor color)
                         | FriendWithFood => is_ant_wf(ant, color)
                         | FoeWithFood => is_ant_wf(ant, othercolor color)
                         | Food => food > 0
                         | Rock => false
                         | (Marker m) => mark_test (marks, (color, m))
                         | FoeMarker => mark_test_any (marks, othercolor color)
                         | Home => hill = SOME color
                         | FoeHome => hill = SOME (othercolor color))
          end

      fun step ant =
        let

          fun set_dir d = 
            (case Array.sub(ants, ant) of
               NONE => raise Step "impossible5"
             | SOME { state, color, resting, dir, has_food, pos, names } =>
                 Array.update(ants, ant, SOME { state = state, color = color,
                                                resting = resting, dir = d,
                                                has_food = has_food, pos = pos,
                                                names = names }))

          fun set_has_food b = 
            (case Array.sub(ants, ant) of
               NONE => raise Step "impossible6"
             | SOME { state, color, resting, dir, has_food, pos, names } =>
                 Array.update(ants, ant, SOME { state = state, color = color,
                                                resting = resting, dir = dir,
                                                has_food = b, pos = pos, names = names }))

          fun adjacent_of_color p c =
            Util.ford 0 5      0
            (fn (d, count) =>
             case get_cell_ant map (next_cell(p, d)) of
               NONE => count
             | SOME aa => 
                 (case Array.sub(ants, aa) of
                    NONE => raise Step "inconsistent"
                  | SOME { color, ... } =>
                      if c = color then count + 1 else count))

          fun check_death p =
            (case get_cell_ant map p of
               NONE => ()
             | SOME aa =>
                 (case Array.sub (ants, aa) of
                    NONE => raise Step "inconsistent2"
                  | SOME { color, has_food, ... } =>
                      if adjacent_of_color p (othercolor color) >= 5
                      then 
                        let in
                          (* death *)
                          Array.update(ants, aa, NONE);
                          set_cell_ant map p NONE;
                          set_cell_food map p (get_cell_food map p + 
                                               3 + (if has_food then 1 else 0));
                          ()
                        end
                      else ()))

          (* check for deaths at p, and in all the adjacent squares *)
          fun check_deaths p =
            let in
              check_death p;
              Util.for 0 5 (fn d => check_death (next_cell(p, d)))
            end
         
        in
          (* if ant is alive *)
          (case Array.sub(ants, ant) of
             NONE => ()
           | SOME { state, color, resting, dir, has_food, pos, names } =>
               if !resting > 0 then resting := (!resting - 1)
               else (case inst color (!state) of
                       Sense(sensedir, st1, st2, cond) => 
                           let 
                               val pos' = sensed_cell (!pos, dir, sensedir)
                               val b = cell_matches (pos', cond, color)
                           in
                               state := (if b then st1 else st2)
                           end
                     (* XXX should check that these indices are in range
                        (otherwise we can set/clear enemy scents!!) *)
                     | Mark(i, st) =>
                         let in
                           set_cell_marks map (!pos)
                            (mark_set (get_cell_marks map (!pos), (color, i)));
                           state := st
                         end
                     | Unmark(i, st) =>
                         let in
                           set_cell_marks map (!pos)
                            (mark_clear (get_cell_marks map (!pos), (color, i)));
                           state := st
                         end
                     | PickUp(st1, st2) =>
                         if has_food orelse get_cell_food map (!pos) = 0
                         then state := st2
                         else
                           let in
                             set_cell_food map (!pos) (get_cell_food map (!pos) - 1);
                             set_has_food true;
                             state := st1
                           end
                     | Drop st =>
                         let in
                           if has_food
                           then set_cell_food map (!pos) (get_cell_food map (!pos) + 1)
                           else ();
                           set_has_food false;
                           state := st
                         end
                     | Turn (lr, st) => 
                         let in
                           set_dir (Ants.turn (lr, dir));
                           state := st
                         end
                       (* XXX have pedantic mode *)
                     | Print s =>
                         let in
                           if (List.exists (fn s => Int.toString ant = s) (!trace))
                           then print (s ^ "\n")
                           else ();
                           state := (!state + 1)
                         end
                     | Abort s =>
                         let in
                           print ("Robot " ^ Int.toString ant ^ " aborted in round " ^ 
                                  Int.toString roundnum ^ " with message:\n" ^
                                  s);
                           raise Step "aborted"
                         end

                     | SetName (c1, c2) =>
                         let in
                           names := (c1, c2);
                           state := (!state + 1)
                         end

                     | Move (st1, st2) =>
                         let
                           val newp = next_cell (!pos, dir)
                         in
                           case get_cell (map, newp) of
                             Rocky => state := st2
                           | Clear { ant = ref (SOME _), ... } => state := st2
                           | Clear { ant = ref NONE, ... } =>
                               let in
                                 set_cell_ant map (!pos) NONE;
                                 set_cell_ant map newp (SOME ant);
                                 pos := newp;
                                 state := st1;
                                 resting := 14;
                                 check_deaths newp
                               end
                         end
                     | Flip (p, st1, st2) => 
                         state := (if Random.randomint p = 0 then st1 else st2)))
        end
    in
      if !vis andalso (roundnum mod (Params.asint 1 freq) = 0)
      then 
        let in
          Vis.showmap ants map roundnum
        end
      else ();
      Util.for 0 (Array.length ants - 1) step
    end

end
