
structure Vis :> VIS =
struct

  open Ants

  exception Vis of string

  datatype gridpiece = Grass | RedHill | BlackHill | Rock


  val color = Params.flag false
			  (SOME ("-color", 
				 "Colorualize every step")) "color"
  val xvis = Params.flag false
	     (SOME ("-xvis", "Xualize every step")) "xvis"

  val xvistext = Params.flag false
			     (SOME ("-xvistext", "Show text on X")) "xvistext"

  val title = "Future Deadly Arena-nitz"

  fun memo f =
      let
	  exception Impossible
	  val r = ref NONE
      in
	  (fn x =>
	      case !r of
		  NONE => let val res = f x
			  in r := SOME res;
	                     res
			  end
		| SOME old => old)
      end

  fun mkxdraw () =
      if !xvis then
	  let
	      val ww = 912
	      val wh = 912
	      val disp = MLX.opendisplay NONE
	      val rt = MLX.defaultrootwindow disp

	      val w = MLX.createsimplewindow disp rt 5 5 ww wh 1 0wx000000 0wx111111

	      val gc = MLX.creategc disp (MLX.wd w)

	      val _ = MLX.mapwindow disp w
	      val _ = MLX.raisewindow disp w
	      val _ = MLX.flush disp


	      fun xdraw x y width height color =
		  let in
		      MLX.setforeground disp gc color;
		      MLX.fillrectangle disp (MLX.wd w) gc x y width height
		  end

	      fun xwrite x y text color =
		  let in
		      MLX.setforeground disp gc color;
		      MLX.drawtext disp (MLX.wd w) gc NONE x y text
		  end

	      fun xclear () =
		  let in
		      MLX.setforeground disp gc 0wx000000;
		      MLX.fillrectangle disp (MLX.wd w) gc 0 0 ww wh
		  end

	      fun xdrawpoint x y color =
		  let in
		      MLX.setforeground disp gc color;
		      MLX.drawpoint disp (MLX.wd w) gc x y 
		  end

	      fun xname name = 
		  let in
		      MLX.storename disp w name
		  end

	  in
	      (* print "Initialized X\n"; *)
	      (xdraw,xwrite,xclear,xdrawpoint,xname)
	  end

      else
	  let
	      fun xdraw _ _ _ _ _ = ()
	      fun xwrite _ _ _ _ = ()
	      fun xclear () = ()
	      fun xdrawpoint _ _ _ = ()	
	      fun xname _ = ()
	  in
	      (xdraw,xwrite,xclear,xdrawpoint,xname)
	  end

  val xfns = memo mkxdraw
  fun xdraw () a b c d e = ( #1 (xfns ()) ) a b c d e
  fun xwrite a b c d = ( #2 (xfns ()) ) a b c d
  fun xclear () = ( #3 (xfns ()) ) ()
  fun xdrawpoint a b c = ( #4 (xfns ()) ) a b c
  fun xname a = ( # 5 (xfns ()) ) a


  fun xdrawblock px py item =
      let 
	  val xoff = if py mod 2 = 0 then 0 else 4
	  val w = 8
	  val h = 8
	  val x = (px * 9) + xoff
	  val y = (py * 9)
		  
      in
	  if !xvistext then xdraw () (x+8) y 1 9 0wx000000 else ();
	  if !xvistext then xdraw () x (y+8) 9 1 0wx000000 else ();
	  (case item of
	       RedHill   => xdraw () x y w h 0wx990000
	     | BlackHill => xdraw () x y w h 0wx666666
	     | Grass     => xdraw () x y w h 0wx005500
	     | Rock      => xdraw () x y w h 0wxca5109)
      end

  fun xdrawant px py ant dir name = 
      let
	  val color = (case ant of 
			   Black => 0wx000000
			 | Red   => 0wxff0000)
	  val xoff = if py mod 2 = 0 then 0 else 4
	  val tx = (9 * px) + xoff
	  val ty = (9 * py)
	  val x = tx + (case (dir + 6) mod 6 of
			    0 => 5
			  | 1 => 5
			  | 5 => 5
			  | 2 => 0
			  | 3 => 0
			  | 4 => 0
			  | _ => 0)
	  val y = ty + (case (dir + 6) mod 6 of
			    5 => 0
			  | 4 => 0
			  | 3 => 3
			  | 0 => 3
			  | 2 => 5
			  | 1 => 5
			  | _ => 0)

	  val nondefaulttext = (name <> #"X") andalso (name <> #"O")
      in
	  if !xvistext andalso nondefaulttext then xwrite (tx+3) (ty-1) (implode [name]) 0wxffffff else ();
	  xdraw () x y 3 3 color
      end

  fun xdrawmark px py marks =
      let
	  val bmark = mark_test_any (marks, Black)
	  val rmark = mark_test_any (marks, Red)

	  val xoff = if py mod 2 = 0 then 0 else 4
	  val w = 2
	  val h = 3
	  val x = (px * 9) + xoff
	  val y = (py * 9) + 1
	  
	  val bitcol = 0wx000000

      in
	  if bmark then let
		  val mx = x + 1
		  val my = y + 3
	      in
		  xdraw () mx my w h 0wxff61ff;
		  if mark_test (marks, (Black, 0)) then xdrawpoint mx my bitcol else ();
		  if mark_test (marks, (Black, 1)) then xdrawpoint mx (my+1) bitcol else ();
		  if mark_test (marks, (Black, 2)) then xdrawpoint mx (my+2) bitcol else ();
		  if mark_test (marks, (Black, 3)) then xdrawpoint (mx+1) my bitcol else ();
		  if mark_test (marks, (Black, 4)) then xdrawpoint (mx+1) (my+1) bitcol else ();
		  if mark_test (marks, (Black, 5)) then xdrawpoint (mx+1) (my+2) bitcol else ()
	      end
	  else ();
	  if rmark then let
		  val mx = x + 5
		  val my = y
	      in
		  xdraw () mx my w h 0wxff9a61;
		  if mark_test (marks, (Red, 0)) then xdrawpoint mx my bitcol else ();
		  if mark_test (marks, (Red, 1)) then xdrawpoint mx (my+1) bitcol else ();
		  if mark_test (marks, (Red, 2)) then xdrawpoint mx (my+2) bitcol else ();
		  if mark_test (marks, (Red, 3)) then xdrawpoint (mx+1) my bitcol else ();
		  if mark_test (marks, (Red, 4)) then xdrawpoint (mx+1) (my+1) bitcol else ();
		  if mark_test (marks, (Red, 5)) then xdrawpoint (mx+1) (my+2) bitcol else ()
	      end
	  else ()
      end

  fun xdrawfood px py amount =
      let
	  val xoff = if py mod 2 = 0 then 0 else 4
	  val w = 2
	  val h1 = if amount > 8 then 8 else amount
	  val h2 = if amount > 8 then if amount > 16 then 8 else amount - 8 else 0
	  val x = (px * 9) + xoff + 2
	  val y = (py * 9)
	  val color = 0wxe6f40e
      in
	  if amount > 0 then xdraw () x (y + 8 - h1) w h1 color else ();
	  if amount > 8 then xdraw () (x+w) (y + 8 - h2) w h2 color else ()
      end



  fun visprint s =
      if not (!xvis) then print s else ()
    
  fun pcolor s =
      if !color then visprint s else ()

  fun showmap ants m round =
    let
      val (w, h) = get_dimensions m
    in
      xname (title ^ " Round: " ^ Int.toString round);
      if round = 0 then visprint "\027[2B" else ();
      visprint "\027[H";
      visprint ("\nBefore round roundnum: " ^ Int.toString round ^ "\n");

      Util.for 0 (h - 1)
      (fn y =>
       let in
         if y mod 2 = 1
         then visprint " "
         else ();
	 
         Util.for 0 (w - 1)
		  (fn x =>
		      let in
			  (case get_cell (m, mkpos (x, y)) of
			       Rocky => (visprint "# ";
					 xdrawblock x y Rock)
			     | Clear { ant = ref ant, food, hill, marks } => 
			       let
			       in
				   (case food of 
					0 => (case hill of
						  (* doesn't currently print marks *)
						      NONE => (visprint ".";
							       xdrawblock x y Grass)
						| SOME Red => (pcolor "\027[0;31;40m";
							       xdrawblock x y RedHill;
							       visprint "+";
							       pcolor "\027[0m")
						| SOME Black => (pcolor "\027[0;32;40m";
								 xdrawblock x y BlackHill;
								 visprint "-";
								 pcolor "\027[0m"))
				      | n =>
					let in
					    (case hill of
						 NONE       => xdrawblock x y Grass
					       | SOME Red   => xdrawblock x y RedHill
					       | SOME Black => xdrawblock x y BlackHill);
					    pcolor "\027[5;34;40m";
					    visprint (implode [String.sub("0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ", n)] handle _ => "!");
					    pcolor "\027[0m"
					end);

				   xdrawmark x y marks;
				   xdrawfood x y food;

				   case ant of 
				       NONE => (pcolor "\027[0;33;40m";
						visprint (implode [mark_to_char marks]);
						pcolor "\027[0m")
				     | SOME a => 
					(case Array.sub(ants, a) of
					     NONE => raise Vis "impossible:inconsistent map"
					   | SOME ({ color = Red, 
						     names=ref(r, b),
						     dir=dir, ... } : Ants.ant) => 
					     (pcolor "\027[0;30;41m";
					      xdrawant x y Red dir r;
					      visprint (implode [r]);
					      pcolor "\027[0m")
					   | SOME ({ names=ref(_, b), dir=dir, ... }) => 
					     (pcolor "\027[0;30;47m";
					      xdrawant x y Black dir b;
					      visprint (implode [b]);
					      pcolor"\027[0m"))

			       end)
		      end);
		  visprint "\n"
       end)
    end
	
end
